/*
 * Broker
 *
 * Allows inter-race ship trading
 *
 * by Roger Burton West <roger@firedrake.demon.co.uk>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tkhost.h>
#include <inifile.h>

const char ver[]="0.2";

int comm;
int premium;
int mintech;
int mindef;
int minfight;

struct BASELIST {
  UB1 racenum;
  Coord pos;
  UB4 money;
  XString fcode;
};

struct BASELIST list_s[NUM_PLANETS+1];
int index_to_s[NUM_PLANETS];
int size_s=0;

struct SHIPLIST{
  UB1 race;
  UB2 location;
  UB4 baseprice;
  UB4 totalprice;
  XString descript;
  UB1 pri_enemy;
  UB2 offerplanet[11];
};

struct SHIPLIST list_a[MAXSHIP+1];
int index_to_a[MAXSHIP];
int size_a=0;

FILE *log;
RACENM *racenmpad;

void dohelp(FILE *stream);
void messoffsucc(int racenum,int shipnum);
void messofffail(int racenum,int shipnum);
void messsold(int owner,int shipnum,int to);
void messoffers(int owner,int shipnum,const char *offers);
void messadvert(int owner,int shipnum);
void messprivate(int owner,int shipnum,int target);
XString racename(int racenum,int mode);
void send_message(int to,XString message);

/* invocation:
broker dir
*/

int main(int argc, char *argv[])
{
  PLANETDATA *planpad;
  BASEDATA *basepad;
  PLANET_COORDS *plcpad;
  SHIPDATA *shippad;
  HULLSPEC *hullpad;
  ENGSPEC *engpad;
  BEAMSPEC *beampad;
  TORPSPEC *torppad;
  XString temp;
  INIFile ini;
  unsigned long price;
  char temp_str[33];
  char temp_str2[33];
  bool conreq[11];
  int k;
  bool v;
  int vv;
  int x;
  int size_b=0;

  if (argc!=2) { // check argument count only
    fprintf(stderr,"Error in invocation!\n");
    dohelp(stderr);
    exit(1);
  }
  set_game_dir(argv[1]);

  if (ini.Open("broker.ini") == RAT_OK) {
    ini.Section("Broker");
    comm=atoi((const char *)(ini.ReadKey("Commission")));
    premium=atoi((const char *)(ini.ReadKey("Premium")));
    mintech=atoi((const char *)(ini.ReadKey("Min_Tech")));
    mindef=atoi((const char *)(ini.ReadKey("Min_Def")));
    minfight=atoi((const char *)(ini.ReadKey("Min_Fighters")));
    ini.Close();
  } else {
    comm=5;
    premium=5;
    mintech=20;
    mindef=10;
    minfight=0;
  }

  log=mkopen("broker.log",false,'o');
  fprintf(log,"Broker v%s running in directory %s\n",ver,argv[1]);

  planpad=read_all_planetdata();
  if (!planpad) {
    fprintf(log,"Couldn't open pdata.hst\n");
    return(1);
  }
  basepad=read_all_basedata();
  if (!basepad) {
    fprintf(log,"Couldn't open bdata.hst\n");
    return(1);
  }
  plcpad=read_all_planet_coords();
  if (!plcpad) {
    fprintf(log,"Couldn't open xyplan.dat\n");
    return(1);
  }
  shippad=read_all_shipdata();
  if (!shippad) {
    fprintf(log,"Couldn't open ships.hst\n");
    return(1);
  }
  hullpad=read_all_hullspec();
  if (!hullpad) {
    fprintf(log,"Couldn't open hullspec.dat\n");
    return(1);
  }
  engpad=read_all_engspec();
  if (!engpad) {
    fprintf(log,"Couldn't open engspec.dat\n");
    return(1);
  }
  beampad=read_all_beamspec();
  if (!beampad) {
    fprintf(log,"Couldn't open beamspec.dat\n");
    return(1);
  }
  torppad=read_all_torpspec();
  if (!torppad) {
    fprintf(log,"Couldn't open torpspec.dat\n");
    return(1);
  }
  racenmpad=read_all_racenm();
  if (!racenmpad) {
    fprintf(log,"Couldn't open race.nm\n");
    return(1);
  }

  for (k=0;k<11;k++) conreq[k]=false;
  v=false;

  for (int p=1;p<=500;p++) {
    if (planpad[p].sCode.Lower()=="con") {
      conreq[planpad[p].wOwner-1]=true;
      v=true;
    }
  }

  if (v) { // at least one config request exists
    temp="(-b0000)<<< Message from BROKER >>>\r\r";
    temp|="Current configuration is:\r";
    temp|="Commission rate: ";
    itoa(comm,temp_str,10);
    temp|=temp_str;
    temp|="%\r";
    temp|="MCr/kt premium: ";
    itoa(premium,temp_str,10);
    temp|=temp_str;
    temp|="\r";
    if (mintech>0) {
      temp|="Minimum tech total: ";
      itoa(mintech,temp_str,10);
      temp|=temp_str;
      temp|="\r";
    }
    if (mindef>0) {
      temp|="Minimum defence total: ";
      itoa(mindef,temp_str,10);
      temp|=temp_str;
      temp|="\r";
    }
    if (minfight>0) {
      temp|="Minimum fighter total: ";
      itoa(minfight,temp_str,10);
      temp|=temp_str;
      temp|="\r";
    }

    for (k=0;k<11;k++) { // send message to all requesting config
      if (conreq[k]) {
        fprintf(log,"Sending config to player %d\n",k+1);
        send_message(k+1,temp);
      }
    }
  }

  /*
   *  make list of qualifying starbases
   */

  k=1;
  for (int b=1;b<501;b++) {
    v=true;
    if (basepad[b].wOwner == 0) v=false;
    if (basepad[b].wDefence < mindef) v=false;
    if (basepad[b].wTechEngine+basepad[b].wTechHull+basepad[b].wTechBeam+basepad[b].wTechTorp < mintech) v=false;
    if (basepad[b].wFighters < minfight) v=false;
    if (v) {
      list_s[k].racenum=basepad[b].wOwner;
      list_s[k].pos=plcpad[b].pos;
      list_s[k].fcode=planpad[b].sCode;
      list_s[k].money=planpad[b].cargo.money+planpad[b].cargo.supplies;
      index_to_s[size_s]=k;
      size_s++;
    }
    k++;
  }

  if (size_s>0) {

    /*
     *  make list of eligible ships
     */

    k=1;
    for (int s=1;s<501;s++) {
      if (shippad[s].wRace >0) {
        v=true;
        // Named $something
        if (shippad[s].sName.GetAt(0) != '$') v=false;
        // Undamaged
        if (shippad[s].wDamage != 0) v=false;
        // Not moving
        if (shippad[s].wWarp != 0) v=false;
        // At qualifying starbase
        if (v) {
          v=false;
          for (int t=0;t<size_s;t++) {
            if (list_s[index_to_s[t]].pos == shippad[s].pos) {
              v=true;
              list_a[k].location=index_to_s[t];
            }
          }
        }
        if (v) {
          fprintf(log,"Race %d offers ship %d for sale.\n",shippad[s].wRace,k);
          list_a[k].pri_enemy=shippad[s].wEnemy;
          list_a[k].race=shippad[s].wRace;

          // construct minimum price and description: hull, eng, beams, torps
          price=hullpad[shippad[s].wShipType].cargo.money+
                hullpad[shippad[s].wShipType].wMass*premium;
          list_a[k].descript="Hull: ";
          list_a[k].descript|=hullpad[shippad[s].wShipType].sName;
          list_a[k].descript|="\r";

          price += engpad[shippad[s].wDriveLevel].cargo.money*
                    hullpad[shippad[s].wShipType].wEngines;
          list_a[k].descript|="Engine";
          if (hullpad[shippad[s].wShipType].wEngines==1) {
            list_a[k].descript|=": ";
          } else {
            list_a[k].descript|="s: ";
            itoa(hullpad[shippad[s].wShipType].wEngines,temp_str,10);
            list_a[k].descript|=temp_str;
            list_a[k].descript|=" x ";
          }
          list_a[k].descript|=engpad[shippad[s].wDriveLevel].sName;
          list_a[k].descript|="\r";

          price += beampad[shippad[s].wBeamLevel].cargo.money*
                    shippad[s].wBeamWeapons;
          switch (shippad[s].wBeamWeapons) {
            case 0:
              list_a[k].descript|="No beam weapons";
              break;
            case 1:
              list_a[k].descript|="Beam: ";
              list_a[k].descript|=beampad[shippad[s].wBeamLevel].sName;
              break;
            default:
              list_a[k].descript|="Beams: ";
              itoa(shippad[s].wBeamWeapons,temp_str,10);
              list_a[k].descript|=temp_str;
              list_a[k].descript|=" x ";
              list_a[k].descript|=beampad[shippad[s].wBeamLevel].sName;
          }
          list_a[k].descript|="\r";

          if (shippad[s].wTorpsBays>0) {
            price += torppad[shippad[s].wTorpLevel].cargo.money*shippad[s].wTorpsBays;
            list_a[k].descript|="Torp";
            if (shippad[s].wTorpsBays==1) {
              list_a[k].descript|=": ";
            } else {
              list_a[k].descript|="s: ";
              itoa(shippad[s].wTorpsBays,temp_str,10);
              list_a[k].descript|=temp_str;
              list_a[k].descript|=" x ";
            }
            list_a[k].descript|=torppad[shippad[s].wTorpLevel].sName;
          } else {
            list_a[k].descript|="No torpedo tubes";
          }
          list_a[k].descript|="\r";

          // get delta-price or enforce minimum
          if (temp.GetAt(1) == '+') {
            list_a[k].baseprice = price + atoi((const char *)temp.Mid(2));
          } else {
            list_a[k].baseprice=atoi((const char *)temp.Mid(1));
            if (price > list_a[k].baseprice) list_a[k].baseprice=price;
          }
          // rename ship appropriately
          itoa(list_a[k].baseprice,temp_str,10);
          temp="$";
          temp|=temp_str;
          shippad[s].sName=temp;
          write_shipdata(s,&shippad[s]);
          // construct price with commission
          list_a[k].totalprice=(long)((double)list_a[k].baseprice*(1.+(double)comm/100.));
          list_a[k].descript|="Asking price: MCr ";
          itoa(list_a[k].totalprice,temp_str,10);
          list_a[k].descript|=temp_str;
          list_a[k].descript|="\r";
          for (x=0;x<11;x++) {
            list_a[k].offerplanet[x]=0;
          }
          index_to_a[size_a]=k;
          size_a++;
        }
        k++;
      }
    }

    fprintf(log,"Finished checking ships for sale.\n");

    if (size_a>0) {

      /*
       *  make list of offers made
       */

      for (int t=size_s-1;t>=0;t--) {
        strcpy(temp_str,(const char *)list_s[index_to_s[t]].fcode);
        for (x=0;x<size_a;x++) {
          itoa(index_to_a[x],temp_str2,10);
          temp=temp_str2;
          while (temp.Len()<3) {
            temp="0"+temp;
          };
          strcpy(temp_str2,temp);
          if (!strcmp(temp_str,temp_str2)) {
            if (list_s[index_to_s[t]].money >= list_a[index_to_a[x]].totalprice) {
              list_a[index_to_a[x]].offerplanet[list_s[index_to_s[t]].racenum-1]=index_to_s[t];
              fprintf(log,"Race %d offers for ship %d.\n",list_s[index_to_s[t]].racenum,index_to_a[x]);
              size_b++;
            }
          }
        }
      }

      fprintf(log,"Finished checking offers made.\n");

      if (size_b>0) {

        /*
         * perform trades
         */
        for (int t=0;t<size_a;t++) {
          vv=list_a[index_to_a[t]].pri_enemy;
          if (vv>0) {
          vv--;
            for (x=0;x<11;x++) {
              if (list_a[index_to_a[t]].offerplanet[x]>0) {
                if (x==vv) {
                  fprintf(log,"Race %d buys ship %d from race %d for %ld/%ld.\n",vv+1,index_to_a[t],list_a[index_to_a[t]].race,list_a[index_to_a[t]].baseprice,list_a[index_to_a[t]].totalprice);
                  int plto=list_a[index_to_a[t]].offerplanet[x];
                  int plfrom=list_a[index_to_a[t]].location;
                  int s=index_to_a[t];
                  messoffsucc(x+1,s);
                  messsold(shippad[s].wRace,index_to_a[t],x+1);
                  shippad[s].pos=list_s[plto].pos;
                  shippad[s].wMission=1;
                  shippad[s].wEnemy=0;
                  shippad[s].wRace=x+1;
                  shippad[s].sName="Newly bought ship";
                  shippad[s].wWarp=0;

                  planpad[plfrom].cargo.fuel+=shippad[s].cargo.fuel;
                  shippad[s].cargo.fuel=0;
                  planpad[plfrom].cargo.money+=shippad[s].cargo.money;
                  shippad[s].cargo.money=0;
                  planpad[plfrom].cargo.clans+=shippad[s].cargo.clans;
                  shippad[s].cargo.clans=0;
                  planpad[plfrom].cargo.tri+=shippad[s].cargo.tri;
                  shippad[s].cargo.tri=0;
                  planpad[plfrom].cargo.dur+=shippad[s].cargo.dur;
                  shippad[s].cargo.dur=0;
                  planpad[plfrom].cargo.mol+=shippad[s].cargo.mol;
                  shippad[s].cargo.mol=0;
                  planpad[plfrom].cargo.supplies+=shippad[s].cargo.supplies;
                  shippad[s].cargo.supplies=0;
                  planpad[plfrom].cargo.money += list_a[index_to_a[t]].baseprice;
                  write_planetdata(plfrom,&planpad[plfrom]);

                  if (list_a[index_to_a[t]].totalprice > planpad[plto].cargo.money) {
                    list_a[index_to_a[t]].totalprice-=planpad[plto].cargo.money;
                    planpad[plto].cargo.money=0;
                    planpad[plto].cargo.supplies-=list_a[index_to_a[t]].totalprice;
                  } else {
                    planpad[plto].cargo.money -= list_a[index_to_a[t]].totalprice;
                  }
                  write_planetdata(plto,&planpad[plto]);

                  write_shipdata(s,&shippad[s]);
                } else {
                  messofffail(x+1,index_to_a[t]);
                }
              }
            }
            index_to_a[t--]=index_to_a[size_a--];
          }
        }

        fprintf(log,"Finished trades.\n");

        // for each ship with offers, tell owner
        for (int t=0;t<size_a;t++) {
          temp="";
          for (x=0;x<11;x++) {
            if (list_a[index_to_a[t]].offerplanet[x]>0) {
              temp |= (char)(65+x);
            }
          }
          if (temp != "") {
            messoffers(list_a[index_to_a[t]].race,index_to_a[t],(const char *)temp);
          }
        }
      }

      // for each ship for sale, tell all
      for (int t=0;t<size_a;t++) {
        if (list_a[index_to_a[t]].pri_enemy==0) {
          messadvert(list_a[index_to_a[t]].race,index_to_a[t]);
        } else {
          messprivate(list_a[index_to_a[t]].race,index_to_a[t],list_a[index_to_a[t]].pri_enemy);
        }
      }
    }
  }

  fprintf(log,"BROKER shutting down.\n\n");
  fclose(log);
  return 0;
}

void dohelp(FILE *stream)
{
  fprintf(stream,"BROKER vv%s - by Roger Burton West <roger@firedrake.demon.co.uk>\n",ver);
  fprintf(stream,"\n");
  fprintf(stream,"Invocation:\n");
  fprintf(stream,"  broker <gamedir> <comm%%> <$/kt> <mintech> <mindefence> <minfighter>\n");
  fprintf(stream,"\n");
  fprintf(stream,"  gamedir: the game directory\n");
  fprintf(stream,"  comm%%: the commission rate\n");
  fprintf(stream,"  $/kt: the price premium per kiloton\n");
  fprintf(stream,"  mintech: minimum total of starbase tech levels\n");
  fprintf(stream,"  mindefence: minimum starbase defence\n");
  fprintf(stream,"  minfight: minimum starbase fighters\n");
  return;
}

void messoffsucc(int racenum,int shipnum)
{
  XString message;
  char temp_str[4];

  message="(-b0001)<<< Message from BROKER >>>\r\r";
  message|="Your bid for the ship\r";
  message|="number ";
  itoa(shipnum,temp_str,10);
  message|=temp_str;
  message|=" was successful.\r";
  send_message(racenum,message);
  return;
}

void messofffail(int racenum,int shipnum)
{
  XString message;
  char temp_str[4];

  message="(-b0002)<<< Message from BROKER >>>\r\r";
  message|="Unfortunately, your bid for the\r";
  message|="ship number ";
  itoa(shipnum,temp_str,10);
  message|=temp_str;
  message|=" was unsuccessful.\r";
  send_message(racenum,message);
  return;
}

void messsold(int owner,int shipnum,int to)
{
  XString message;
  char temp_str[4];

  message="(-b0003)<<< Message from BROKER >>>\r\r";
  message|="Your ship number ";
  itoa(shipnum,temp_str,10);
  message|=temp_str;
  message|=" has been sold to\r";
  message|=racenmpad[to].sFull;
  message|=".\r";
  send_message(owner,message);
  return;
}

void messoffers(int owner,int shipnum,const char *offers)
{
  XString message;
  XString off;
  char temp_str[4];

  off=offers;
  message="(-b0004)<<< Message from BROKER >>>\r\r";
  if (off.Len()==1) {
    message|="We have received an offer from\r";
    message|=racenmpad[off.GetAt(0)-64].sFull;
  } else {
    message|="We have received offers from:\r";
    for (int x=0;x<off.Len();x++) {
      message|=racenmpad[off.GetAt(x)-64].sFull;
      message|="\r";
    }
  }
  message|="\r";
  message|="for your ship number ";
  itoa(shipnum,temp_str,10);
  message|=temp_str;
  message|=".\r";
  send_message(owner,message);
  return;
}

void messadvert(int owner,int shipnum)
{
  XString message,temp;
  char temp_str[4];
  int x;

  message="(-b0005)<<< Message from BROKER >>>\r\r";
  message|=racenmpad[owner].sFull;
  message|="\roffers a ship for sale:\r\r";
  message|=list_a[shipnum].descript;
  itoa(shipnum,temp_str,10);
  temp=temp_str;
  while (temp.Len()<3) {
    temp = "0" + temp;
  }
  message|="\rQuote reference # ";
  message|=temp;
  message|="\r";
  for (x=1;x<12;x++) {
    if (x!=owner) send_message(x,message);
  }
  return;
}

void messprivate(int owner,int shipnum,int target)
{
  XString message,temp;
  char temp_str[4];

  message="(-b0006)<<< Message from BROKER >>>\r\r";
  message|=racenmpad[owner].sFull;
  message|="\roffers a ship for private sale to\r";
  message|=racenmpad[target].sFull;
  message|=":\r\r";
  message|=list_a[shipnum].descript;
  itoa(shipnum,temp_str,10);
  temp=temp_str;
  while (temp.Len()<3) {
    temp = "0" + temp;
  }
  message|="\rQuote reference # ";
  message|=temp;
  message|="\r";
  send_message(target,message);
  return;
}

void send_message(int to,XString message)
{
  OUTMESS temp;
  temp.iTo=to;
  temp.sMessage=message;
  write_outmess(0,&temp);
}
